         subroutine eigen(n, nvec, a, lda, w, z, ldz, mf, mb)
         use OMP_LIB
         implicit NONE
*-
         integer, intent(in)    :: n, nvec, lda, ldz, mf, mb
         real(8), intent(inout) :: a(lda,*), w(*), z(ldz,*)
         real(8), pointer       :: d(:), e(:), e2(:)
*-
         real(8)                :: hs0, hs1, s0, s1, r, ret
         integer                :: my_rank, world_size
         integer(8)             :: INFO
         integer                :: ierr, i
*-
         integer(8)             :: buffersize
         integer(8), external   :: eigen_buffersize
         include 'trd.h'
*-
#if USE_TIMER
         hs0 = OMP_Get_wtime()
#endif
         call CUBLAS_INIT()

         allocate(d(1:n), e(1:n), e2(1:n))

         buffersize = eigen_buffersize(n, nvec, lda, ldz, mf, mb)
         ierr = cuda_alloc( devPtr_A, buffersize )
         size_A = (buffersize+7)/8
         if ( ierr == 1 ) then
!           print*, "Memory allocation fault on GPU [eigen]"
           devPtr_A = 0_8
           size_A = 0_8
         endif
*-
         call cuda_sync(1)
#if USE_TIMER
         s0 = OMP_Get_wtime()
#endif

!         call tred1(n, a(1,1), lda, d(1), e(1), mf)
         call hhsy2tr(n, a(1,1), lda, d(1), e(1), mf)
!         print*,"D=",D(1:n)
!         print*,"E=",E(1:n)

         call cuda_sync(1)
#if USE_TIMER
         s1 = OMP_Get_wtime()
#endif

#if USE_TIMER
         print*,"TRD-BLK",n," ",s1-s0,
     &            1*dble(n)**3*4/3/(s1-s0)*1D-9,"GFLOPS"
!         print*,"TRD-BLK-INFO",n," ",mf
         call flush(6)
#endif
*-
#if USE_TIMER
         s0 = OMP_Get_wtime()
#endif

!         call xdgemm( "0", "0", 1, 1, 1, 0d0, e2,1, e2,1, 1d0, e2,1);

         e2(1:n) = e(1:n)
         w(1:n)  = d(1:n)

         call dc(n, w(1), e2(2), z(1,1), ldz, INFO, ret)
!         call cuda_sync(1)
!         call xdgemm( "1", "0", 1, 1, 1, 0d0, e2,1, e2,1, 1d0, e2,1);
!         print*,e2(1:3)

         call cuda_sync(1)
#if USE_TIMER
         s1 = OMP_Get_wtime()
         print*,"D&C    ",n," ",s1-s0,1D-9*ret/(s1-s0),'GFLOPS'
         call flush(6)
#endif
*-
#if USE_TIMER
         s0 = OMP_Get_wtime()
#endif

!         call trbakwyx(n, nvec, a, lda, z, ldz, d, e, mb)
         call hhtr2sy(n, nvec, a, lda, z, ldz, mb)

         call cuda_sync(1)
#if USE_TIMER
         s1 = OMP_Get_wtime()
#endif

#if USE_TIMER
         r = 2*dble(nvec)*dble(n)**2
         print*,"TRDBAK ",n," ",s1-s0,1D-9*r/(s1-s0),"GFLOPS"
         call flush(6)
#endif
*-
         deallocate(d, e, e2)
         if ( size_A /= 0 ) then
            call cuda_free( devPtr_A )
         endif
         call CUBLAS_SHUTDOWN()

#if USE_TIMER
         hs1 = OMP_Get_wtime()
#endif
*
#if USE_TIMER
         r = dble(n)**3*4/3 + ret + 2*dble(nvec)*dble(n)**2
         print*,"Total  ",n," ",hs1-hs0,1D-9*r/(hs1-hs0),"GFLOPS"
         call flush(6)
#endif
!         call oo

         return
*-
9999     continue
*-
         end subroutine

