       subroutine ev_test_2D(n, a, nma, w, z, nmz, ret)
       use OMP_LIB
       implicit NONE
       integer                :: n, nma, nmz
       real(8)                :: a(*), w(*), z(*), ret

       call ev_check_2D_body(n, a(1), nma, w(1), z(1), nmz, ret)

       return
       end
*
*
*
       subroutine ev_check_2D_body(n, a, nma, w, z, nmz, ret)
       use OMP_LIB
       implicit NONE

       integer                :: n, nma, nmz
       real(8)                :: a(1:nma,*)
       real(8)                :: w(1:n)
       real(8)                :: z(1:nmz,*)
       real(8)                :: ret

       real(8), allocatable   :: e(:)
       real(8), allocatable   :: f(:)

       integer                :: i, j, k, ldd, ierr

       real(8)                :: r, rr
       real(8)                :: result(10)

       integer(8)             :: l_array, idx
       real(8)                :: m_epsilon
       real(8), external      :: machine_epsilon
       real(8), external      :: ddot
       real(8), external      :: dlamch
       real(8), external      :: dlansy
*
       include 'trd.h'
*-
       m_epsilon = machine_epsilon()
*-
       do i=1,n
          if ( isNAN( w(i) ) ) then
             print*,"W(",i,") :: is detected as NAN"
             stop
          end if
       end do
*-
       do i=1,n
          do j=1,n
             if ( isNAN( z(j,i) ) ) then
                print*,"Z(",j,",",i,") :: is detected as NAN"
                stop
             end if
          end do
       end do
*-
       ldd = MAX(nma, nmz)
       l_array = ldd * n
       allocate(e(1:l_array), f(1:n), stat=ierr)
       if ( ierr /= 0 ) then
          print*,"Shortage of the memory alloction for verification." 
          return
       end if
*
!       size_A = l_array + int(max(nma,nmz),8)*n + int(nmz,8)*n
       i = cuda_alloc_available( devPtr_A, 0_8, size_A )
       size_A = size_A/8
       if ( i /= 0 ) then
          size_A = 0
       end if
*-
!
! E(::) = ZW
! result(4) := | A |_F
! result(1) := | AZ - ZW |_F
!
!$OMP PARALLEL DO PRIVATE(i,j,idx,r,rr)
       do i=1,n
          r = 0D0
          do j=1,n
             rr = z(j,i)*w(i)
             idx = j + INT(i-1,8)*ldd
             e(idx) = rr
             rr = a(j,i)
             r = r + rr**2
          end do
          f(i) = r
       end do
!$OMP END PARALLEL DO
       r = 0D0
!$OMP SIMD
       do i=1,n
          r = r + f(i)
       end do
       result(4) = SQRT(r)

       call xDGEMM( 'N', 'N', n, n, n,
     &            -1D0, a, nma, z, nmz, 1d0, e, ldd)
!$OMP PARALLEL DO PRIVATE(i,j,idx,r,rr)
       do i=1,n
          r = 0D0
!$OMP SIMD
          do j=1,n
             idx = j + INT(i-1,8)*ldd
             rr = e(idx)
             r = r + rr**2
             rr = z(j,i)*w(i)
             e(idx) = rr
          end do
          f(i) = r
       end do
!$OMP END PARALLEL DO
       r = 0D0
!$OMP SIMD
       do i=1,n
          r = r + f(i)
       end do
       result(1) = SQRT(r)/(N*result(4))
*-
!
! result(2) := | A- ZWZ^t |_F
!
       call DGEMM( 'N', 'T', n, n, n,
     &            -1D0, e, ldd, z, nmz, 1d0, a, ldd)
!$OMP PARALLEL DO PRIVATE(i)
       do i=1,n
          f(i) = DDOT( n, a(1,i), 1, a(1,i), 1 )
       end do
!$OMP END PARALLEL DO
       r = 0D0
!$OMP SIMD
       do i=1,n
          r = r + f(i)
       end do
       result(2) = SQRT(r)/(N*result(4))
*-
!
! result(3) := | Z^tZ - I |_F
!
       call DGEMM( 'T', 'N', n, n, n,
     &             1D0, z, nmz, z, nmz, 0D0, e, ldd)
!$OMP PARALLEL DO PRIVATE(i,idx)
       do i=1,n
          idx = i + INT(i-1,8)*ldd
          e(idx) = e(idx) - 1D0
          idx = 1 + INT(i-1,8)*ldd
          f(i) = DDOT( n, e(idx), 1, e(idx), 1 )
       end do
!$OMP END PARALLEL DO
       r = 0D0
!$OMP SIMD
       do i=1,n
          r = r + f(i)
       end do
       result(3) = sqrt(r)/n
*-
       print*,"machine.EPS            =", m_epsilon
       print*,"|A|_F                  =", result(4)
       print*,"|AZ-ZW|_F/(N|A|_F)     =", result(1)
       print*,"|A-ZWZ^t|_F/(N|A|_F)   =", result(2)
       print*,"|Z^tZ-I|_F/(N)         =", result(3)
*
       if ( result(1) > 1D-1 ) then
          stop
       end if
*
       deallocate(e)
       deallocate(f)
*
       if ( size_A /= 0 ) then
          call cuda_sync( 1 )
          call cuda_free( devPtr_A )
       end if
*-
       return
       end
