      program EIGENG_TEST
      use omp_lib
      implicit NONE
!
      real(8), pointer ::   a(:),z(:),w(:)
      real(8), parameter :: ONE = 1D0
      integer :: n, mf, mb, nall, mtype, nx, nm
      integer :: i, j, istat
      real(8) :: x, ax, t
      real(8) :: PAI, theta, ts, te
      integer(8) :: larray
      integer :: gpu_id
      character gpu_id_string*256, fname*256
!
      include 'trd.h'
!
      PAI = 4*ATAN(ONE)
!
      call getenv( "ASPEN_GPU_ID", gpu_id_string )
      if ( gpu_id_string(1:1) /= ' ' ) then
          read(gpu_id_string,*) gpu_id
      else
          gpu_id = 0
      end if
      call cuda_set_device( gpu_id )
      call ASPEN_INIT( gpu_id )


      DO
*
         read(5,*) n, mf, mb, nall, mtype
         print*, n, mf, mb, nall, mtype
         if ( n < 1 ) EXIT
         mb = 256
*
         nx = n
         call CSTAB_get_optdim(nx, 6, 16*4, 16*4*2, nm)
         PRINT*,"N=",n, "NM=",nm, "MF=",mf, "MB=",mb
*
         larray = int(nm,8)*int(n,8)
         allocate(
     &               a(larray), z(larray),
     &               w(n),
     &               stat=istat)
         if(istat.ne.0) then
            print*,"Memory exhausted"
            call flush(6)
            EXIT
         end if
*
         call mat_set(n, a(1), nm, mtype)
*-
         ts = omp_get_wtime()
         call eigen(n, n, a(1), nm, w(1), z(1), nm, mf, mb)
         te = omp_get_wtime()
         print*,"elapsed time (",N,")=",te-ts,"[sec]"
*-
         if(IAND(nall,4)/=0.and.mtype==0)then
            ax=0D0
            do i=1,n
               j=n-i
               theta=PAI*(2*j+1)/(2*n+1)
               x=5D-1/(1D0-COS(theta))
               ax=MAX(ax,ABS(w(i)-x)/ABS(x))
            end do
            print*,"max|w(i)-w(i).true|/|w|=",ax
         end if
         call flush(6)
*-
         if(IAND(nall,4)/=0)then
            call mat_set(n, a(1), nm, mtype) !, NDIM)
            ts = omp_get_wtime()
            call ev_test_2D(n, a(1), nm, w(1), z(1), nm, t)
            te = omp_get_wtime()
            print*,"verifing time =",te-ts,"[sec]"
         end if
*
         deallocate(a)
         deallocate(z)
         deallocate(w)
*
         call sleep(1)
*
      END DO


      call sleep(2)
      close(10)
      call sleep(2)
      call ASPEN_shutdown()
      call sleep(2)

      end program EIGENG_TEST

      subroutine flush(i)
      integer, intent(in) :: i
      end
