#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <cuda_runtime.h>
#include <cuda.h>
#include <cublas.h>

extern "C"
cudaStream_t ASPEN_get_Stream(void);


//---::---//---::---//---::---//---::---//---::---//---::---//---::---
__global__ void 
zero_triangle_gpu_kernel ( int const m, double * x, int const ldx )
{
  const size_t  pos = threadIdx.x;
  x += pos * ldx;
  for(int i=pos;i<m;i++){ x[i] = 0.; }
}

extern "C" void
zero_triangle_gpu ( int const m, double *x, int const ldx, cudaStream_t const stream )
{
  zero_triangle_gpu_kernel <<< 1, m, 0, stream >>> ( m, x, ldx );
}

extern "C" void
zero_triangle_gpu_ ( const int *m, double *dev_x, const int *ldx )
{
  cudaStream_t stream = ASPEN_get_Stream();
  double *x = (double *)(*((unsigned long *)dev_x));
  zero_triangle_gpu_kernel <<< 1, *m, 0, stream >>> ( *m, x, *ldx );
}
//---::---//---::---//---::---//---::---//---::---//---::---//---::---



//---::---//---::---//---::---//---::---//---::---//---::---//---::---
__global__ void 
mul_diag_gpu_kernel ( double * const x,  int const incx )
{
  size_t const pos = incx*threadIdx.x;
  double const t   = x[pos];
  x[pos] = ( t == 0.0 ? 2.0 : t ) * 0.5;
}

extern "C" void
mul_diag_gpu ( int const m, double *x, int const incx, cudaStream_t const stream )
{
  mul_diag_gpu_kernel <<< 1, m, 0, stream >>> ( x, incx );
}

extern "C" void
mul_diag_gpu_ ( const int *m, double *dev_x, const int *incx )
{
  cudaStream_t stream = ASPEN_get_Stream();
  double *x = (double *)(*((unsigned long *)dev_x));
  mul_diag_gpu_kernel <<< 1, *m, 0, stream >>> ( x, *incx );
}
//---::---//---::---//---::---//---::---//---::---//---::---//---::---

