#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <cuda_runtime.h>
#include <cuda.h>
#include <cublas.h>

#include "misc_gpu.hpp"
#include "task_gpu.hpp"


extern "C"
cudaStream_t ASPEN_get_Stream(void);

#define        u_x(x)                u_x[(x)-1]
#define        v_x(x)                v_x[(x)-1]
#define        ST(addr,val)          Store_Global_with_Confirm((addr), (val))
//#define        ST(addr,val)          *(addr) = (val)

__device__ __noinline__ void
t3xx_gpu_sub0( const int global_id, const int i, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, const double beta, double * s0, double * s1 )
{
   const int myid     = threadIdx.x + 1;
   extern __shared__ double shmem[];

   double * const ub  = u_x + ldu;
   double * const vb  = v_x + ldv;
/*
   Scal(l, beta, &v(1), 1);
   const T gamma = Dot(l, &u(1), 1, &v(1), 1) * beta * HALF;
   Axpy(l, gamma, &u(1), 1, &v(1), 1);
*/

   double gamma = 0.;
   {
      double * vbk = vb + myid-1;
      double * ubk = ub + myid-1;
      { int k_, k; const int i_ = i / blockDim.x;
         for(k_=0, k=myid; k_<i_; k_++, k+=blockDim.x) {
            const double t = beta * (*vbk);
            gamma += t * (*ubk);
            vbk += blockDim.x; ubk += blockDim.x;
         } if ( k <= i ) {
            const double t = beta * (*vbk);
            gamma += t * (*ubk);
         }
      }
   }sum_over_warp_masters(gamma);
   gamma *= beta; gamma /= 2;
   {
      double * vbk = vb + myid-1;
      double * ubk = ub + myid-1;
      { int k_, k; const int i_ = i / blockDim.x;
         for(k_=0, k=myid; k_<i_; k_++, k+=blockDim.x) {
            const double t = beta * (*vbk);
            ST ( vbk, fma(gamma, (*ubk), t) );
            vbk += blockDim.x; ubk += blockDim.x;
         } if ( k <= i ) {
            const double t = beta * (*vbk);
            ST ( vbk, fma(gamma, (*ubk), t) );
         }
      }
   } __syncthreads(); // to allow vb(i) from other threads
   {
      double * vbi = vb + (i-1) + (myid-1)*ldv;
      double * ubi = ub + (i-1) + (myid-1)*ldu;
      { int mode_, mode; const int mdone_ = mdone / blockDim.x;
         for(mode_=0, mode=myid; mode_<mdone_; mode_++, mode+=blockDim.x) {
            ST ( &s0[mode-1], (*ubi) );
            ST ( &s1[mode-1], (*vbi) );
            ubi += ldu*blockDim.x; vbi += ldv*blockDim.x;
         } if ( mode <= mdone ) {
            ST ( &s0[mode-1], (*ubi) );
            ST ( &s1[mode-1], (*vbi) );
         }
      }
   }
}

__device__ __noinline__ void
t3xx_gpu_sub1( const int global_id, const int i, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, double * s0, double * s1 )
{
   const int myid     = threadIdx.x + 1;
   const int bcast_id = (get_shmem_size(1)/sizeof(double))-2;
   extern __shared__ double shmem[];

   double * const ub  = u_x + ldu;
   double * const vb  = v_x + ldv;
/*
   if (mdone > 0) {
      T * const ub_dev = &uu(1,m+1);
      T * const vb_dev = &vv(1,m+1);
      cublasDgemv('N', i, mdone, ONE, &ub(1), ldu, &vb(i), ldv, ONE, &u(1), 1);
      cublasDgemv('N', i, mdone, ONE, &vb(1), ldv, &ub(i), ldu, ONE, &u(1), 1);
   }
*/
   const int k  = myid + global_id*blockDim.x;
   const int kk = min(k,i);

   const double * ubk = ub + (kk-1);
   const double * vbk = vb + (kk-1);

   double s = u_x(kk);
   const int mstep = (bcast_id>>1);
   for(int m_=0; m_<mdone; m_+=mstep) {
      const int mmx = min(mstep, mdone-m_);
      __syncthreads();
      if ( myid <= mmx ) {
         const int mode = m_+myid;
         shmem[(myid-1)+0*mstep] = s0[mode-1];
         shmem[(myid-1)+1*mstep] = s1[mode-1];
      } __syncthreads();
      for(int mm=1; mm<=mmx; mm++) {
         s += (*ubk) * shmem[(mm-1)+1*mstep]; ubk += ldu;
         s += (*vbk) * shmem[(mm-1)+0*mstep]; vbk += ldv;
      }
   }
   if ( k <= i ) { ST ( &u_x(k), s ); }
}

__device__ __noinline__ void
t3xx_gpu_sub2( const int global_id, const int i, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, double * beta_, double * d_i, double * e_i )
{
   const int myid     = threadIdx.x + 1;
   const int bcast_id = (get_shmem_size(1)/sizeof(double))-2;
   extern __shared__ double shmem[];

   const int L = i-1;
/*
   anorm = dnrm2( L, u_x, 1 )
   ul    = ZERO
   beta  = ZERO
   if ( anorm > ul ) then
      ul    =  u_x(L)
      anorm = -SIGN(anorm, ul)
      ul    =  ul - anorm
      beta  =  ul / anorm
      u_x(1:L) = u_x(1:L) / ul
      ul    =  ONE
   endif

   u_x(L) =  ul
   e  (i) =  anorm
   d  (i) =  u_x(i)
*/

   double anorm = 0.;
   double delta = 0.;
   {
      for(int k=myid; k<=L; k+=blockDim.x) {
         const double c = u_x(k);
         const double ax = fma(c, c, anorm);
         anorm -= ax;
         const double dx = fma(c, c, anorm);
         delta += dx;
         anorm = ax;
      }
   }delta /= 2; anorm += delta;
   const bool cond_L = ( (myid-1) == (L-1) % blockDim.x );
   if ( cond_L ) {
      shmem[bcast_id] = u_x(L);
   }sum2_over_warp_masters(anorm,delta);
   anorm = sqrt(anorm+delta);

   double uL   = 0.;
   double beta = 0.;
   if ( anorm > 0. ) {
      uL    = shmem[bcast_id];
      anorm = uL >= 0. ? -anorm : +anorm;
      uL    = uL - anorm;
      beta  = uL / anorm;
      {
         const double urec = 1.0 / uL;
         double * uk = &u_x(myid);
         { int k_, k; const int L1_ = (L-1) / blockDim.x;
            for(k_=0, k=myid; k_<L1_; k_++, k+=blockDim.x) {
               ST ( uk, (*uk) * urec ); uk += blockDim.x;
            } if ( k <= L-1 ) {
               ST ( uk, (*uk) * urec );
            }
         }
      }
      uL     = 1.;
   }

   if ( myid == 1 ) {
      ST ( beta_,   beta   );
      ST ( e_i,     anorm  );
      ST ( d_i,     u_x(i) );
      ST ( &u_x(L), uL     );
      ST ( &u_x(i), beta   );
   }
}

__device__ __noinline__ void
t3xx_gpu_sub3( const int global_id, const int L, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, double * s0, double * s1 )
{
   const int myid     = threadIdx.x + 1;
   extern __shared__ double shmem[];

   double * const ub  = u_x + ldu;
   double * const vb  = v_x + ldv;

   const int mode = global_id + 1;
   const double * ubm = ub + (myid-1) + ldu * (mode-1);
   const double * vbm = vb + (myid-1) + ldv * (mode-1);

   double r0 = 0.;
   double r1 = 0.;
   { int k_, k; const int L_ = L / blockDim.x;
      for(k_=0, k=myid; k_<L_; k_++, k+=blockDim.x) {
         const double t = u_x(k);
         r0 += t * (*vbm);
         r1 += t * (*ubm);
         vbm += blockDim.x; ubm += blockDim.x;
      } if ( k <= L ) {
         const double t = u_x(k);
         r0 += t * (*vbm);
         r1 += t * (*ubm);
      }
   } sum2_over_warp_masters(r0, r1);
   if ( myid == 1 ) {
      ST ( &s0[mode-1], r0 );
      ST ( &s1[mode-1], r1 );
   }
}

__device__ __noinline__ void
t3xx_gpu_sub4( const int global_id, const int L, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, const double * s0, const double * s1 )
{
   const int myid     = threadIdx.x + 1;
   const int bcast_id = (get_shmem_size(1)/sizeof(double))-2;
   extern __shared__ double shmem[];

   double * const ub  = u_x + ldu;
   double * const vb  = v_x + ldv;

   const int k  = myid + global_id*blockDim.x;
   const int kk = min(k, L);

   const double * ubk = ub + (kk-1);
   const double * vbk = vb + (kk-1);

   double r = 0.;
   const int mstep = (bcast_id>>1);
   for(int m_=0; m_<mdone; m_+=mstep) {
      const int mmx = min(mstep, mdone-m_);
      __syncthreads();
      if ( myid <= mmx ) {
         const int mode = m_+myid;
         shmem[(myid-1)+0*mstep] = s0[mode-1];
         shmem[(myid-1)+1*mstep] = s1[mode-1];
      } __syncthreads();
      for(int mm=1; mm<=mmx; mm++) {
         r += shmem[(mm-1)+0*mstep] * (*ubk);
         r += shmem[(mm-1)+1*mstep] * (*vbk);
         ubk += ldu; vbk += ldv;
      }
   }
   if ( k <= L ) { ST ( &v_x(k), r ); }
}

__global__ void
t3xx_gpu( const uint4 key, TB_control_t * TB_device, const int i, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, double * d_i, double * workspace )
{
   extern __shared__ double shmem[];

   int global_id = 0;

   double * const e_i   = d_i + ldu;
   double * const s0    = workspace;
   double * const s1    = workspace + mdone;
   double * const beta_ = workspace - 1;


   if ( mdone > 0 ) {

      { TASK_LOOP ( 0, 1 ) {
         const double beta = * beta_;
         t3xx_gpu_sub0( global_id, i, mdone, u_x, ldu, v_x, ldv, beta, s0, s1 );
      } TASK_WAIT ( 0, 1, x ); }


      { const int IX = (i-1)/blockDim.x+1;
      TASK_LOOP ( 1, IX ) {
         t3xx_gpu_sub1( global_id, i, mdone, u_x, ldu, v_x, ldv, s0, s1 );
      } TASK_WAIT ( 1, IX, y ); }

   }

   { TASK_LOOP ( 2, 1 ) {
      t3xx_gpu_sub2( global_id, i, mdone, u_x, ldu, v_x, ldv, beta_, d_i, e_i );
   } TASK_WAIT ( 2, 1, z ); }


   if ( mdone > 0 ) {
      const int L = i-1;

      { TASK_LOOP ( 3, mdone ) {
         t3xx_gpu_sub3( global_id, L, mdone, u_x, ldu, v_x, ldv, s0, s1 );
      } TASK_WAIT ( 3, mdone, w ); }


      { const int LX = (L-1)/blockDim.x+1;
      TASK_LOOP ( 4, LX ) {
         t3xx_gpu_sub4( global_id, L, mdone, u_x, ldu, v_x, ldv, s0, s1 );
      } TASK_WAIT ( 4, LX, x ); }

   }


//   TASK_RESET(5);
}


static int    THREADS_dynamic = 0;
static int    BLOCKS_dynamic  = 0;
static size_t SHMEM_required  = 0;

extern "C" void
t3xx( const int i, const int mdone, double * u_x, const int ldu, double * v_x, const int ldv, double * d_i, double * e_i, double * workspace, void * TB_, cudaStream_t stream )
{
   const int THREADS_max = (32*8);

   if ( THREADS_dynamic <= 0 ) {
      int numBlocks;
      THREADS_dynamic = THREADS_max;
      while ( true ) {
         SHMEM_required = get_shmem_size(THREADS_dynamic);
         cudaOccupancyMaxActiveBlocksPerMultiprocessor( &numBlocks, t3xx_gpu,
                         THREADS_dynamic,
                         SHMEM_required );
         if ( numBlocks > 0 || THREADS_dynamic <= 0 ) break;
         THREADS_dynamic -= 32;
      }

      struct cudaDeviceProp deviceProp;
      cudaGetDeviceProperties ( &deviceProp, 0 );
      BLOCKS_dynamic = deviceProp.multiProcessorCount * numBlocks;
   }


   TB_control_t * TB_device = (TB_control_t *)TB_;
   static uint4 last_key = (uint4){ 0, 0, 0, 0 };
   const uint4 key = gen_random_keys( last_key );

   const int BLOCKS = min( BLOCKS_dynamic,
                   max( (i-1)/THREADS_dynamic+1, mdone ) );

   cudaMemsetAsync( TB_device, 0, sizeof(TB_control_t)*5, stream );
   t3xx_gpu <<< BLOCKS, THREADS_dynamic, SHMEM_required, stream >>>
      ( key, TB_device,
        i, mdone, u_x, ldu, v_x, ldv, d_i, workspace );
}

#undef	u_x
#undef	v_x

