       subroutine tred1_vo(
     &               u_x, v_x,
     &               ux, vx, nv,
     &               u_t, v_t, beta,
     &               i, i_base, m0)
       use OMP_LIB
       implicit none
!
       integer, intent(in)    ::  nv, i, i_base, m0
       real(8), intent(in)    ::  u_x(*)
       real(8), intent(out)   ::  v_x(*)
       real(8), intent(in)    ::  ux(1:nv,*)
       real(8), intent(in)    ::  vx(1:nv,*)
       real(8), intent(inout) ::  u_t(*)
       real(8), intent(inout) ::  v_t(*)
       real(8), intent(in)    ::  beta
!
       integer                ::  j, k, L, n
       integer                ::  k_1, k_2, k_3
!
       integer                ::  i_1, i_2, i_3, i_4
       integer                ::  j_1, j_2, j_3, j_4
       integer                ::  l_1, l_2, l_3, l_4
       integer                ::  jj_1, jj_2, jj_3, jj_4
       integer                ::  LX
!
       include 'trd.h'
       include 'param.h'
       include 'CSTAB.h'
!
       real(8)                ::  w0
       real(8)                ::  u0, v0
       real(8)                ::  u1, v1
       real(8)                ::  u2, v2
       real(8)                ::  ux0, vx0
       real(8)                ::  ux1, vx1
       real(8)                ::  ux2, vx2
!
       real(8)                ::  prod_uv
       real(8), save          ::  alpha
       integer                ::  x_pos, x_owner_nod
       integer                ::  y_pos, y_owner_nod
       integer(8)             ::  idx_VX
       integer(8)             ::  mlen

       real(8), external      ::  ddot
       real(8)                ::  ts, te
       real(8)                ::  t0, t1


       t0 = omp_get_wtime()

       k_1 = i - i_base
       k_2 = m0

       L = i-1
       n = L

       if ( beta /= ZERO ) then
!
! all threads pass
!
          if ( k_2 > k_1 ) then
             ts = omp_get_wtime()
!
! v=v-(UV+VU)u
! here, [u_t, v_t] = [V*u, U*u]
!
             if ( omp_get_thread_num() == 
     $            MIN( 1, omp_get_num_threads()-1  ) ) then
                call dgemv( 'T', n, k_2-k_1,
     $                       1D0, vx(1, k_1+1), nv,
     $                            u_x(1), 1,
     $                       0D0, u_t(1), 1 )
             end if
             if ( omp_get_thread_num() == 
     $            MIN( 2, omp_get_num_threads()-1  ) ) then
                call dgemv( 'T', n, k_2-k_1,
     $                       1D0, ux(1, k_1+1), nv,
     $                            u_x(1), 1,
     $                       0D0, v_t(1), 1 )
             end if
!
! all threads pass
!
!************************************************************
! HERE needs BARRIER to guarrantee the consistency of 'v_x'
!$OMP MASTER
             call cuda_sync( -1 )
!$OMP END MASTER
!$OMP BARRIER
             te = omp_get_wtime()
!$OMP MASTER
             ttt6(4) = ttt6(4) + (te - ts)
!$OMP END MASTER
             ts = omp_get_wtime()
!************************************************************
!
! all threads pass
!
! here, v_x = v_x - Ux*u_t - Vx*v_t

             j_2 = 1; j_3 = n
             l_1 = (n-1)/omp_get_num_threads()+1
             l_1 = ((l_1-1)/64+1)*64
             j_2 = 1 + l_1 * omp_get_thread_num()
             j_3 = MIN(n, j_2 + l_1 - 1)

             if ( j_2 <= j_3 ) then

                call dgemv( 'N', j_3-j_2+1, k_2-k_1,
     $                     -1D0, ux(j_2, k_1+1), nv,
     $                           u_t(1), 1,
     $                      1D0, v_x(j_2), 1 )
                call dgemv( 'N', j_3-j_2+1, k_2-k_1,
     $                     -1D0, vx(j_2, k_1+1), nv,
     $                           v_t(1), 1,
     $                      1D0, v_x(j_2), 1 )
             end if

!
! all threads pass
!
!$OMP BARRIER
             te = omp_get_wtime()
!$OMP MASTER
             ttt6(3) = ttt6(3) + (te - ts)
!$OMP END MASTER
          else
!$OMP MASTER
             call cuda_sync( -1 )
!$OMP END MASTER
          end if

          if ( omp_get_thread_num() == 0 ) then
             ts = omp_get_wtime()

             prod_uv = ddot( n, v_x, 1, u_x, 1 )
             alpha   = prod_uv/(2*beta)
!DIR$ IVDEP
!DIR$ VECTOR ALWAYS
!$OMP SIMD             
             do j_1=1,n
                v_x(j_1) = (v_x(j_1)-alpha*u_x(j_1))/beta
             end do! j_1
             v_x(i:i_base+m0) = ZERO
!
! this corresponds to the barrier before cuda_set_vector in t6
!
! BARRIER ID 0
!$OMP BARRIER
             te = omp_get_wtime()
             ttt6(2) = ttt6(2) + (te - ts)
             ts = omp_get_wtime()

             idx_VX = int(i-i_base-1,8)*int(nv*8,8)
             mlen = int(i_base+m0, 8)
             call cuda_set_vector( mlen, eSize, v_x, devPtr_VX+idx_VX )
!
! this corresponds to the barrier after cuda_set_vector in t6
!
! BARRIER ID 1
!$OMP BARRIER
             te = omp_get_wtime()
             ttt6(1) = ttt6(1) + (te - ts)

          endif

       else

         if ( omp_get_thread_num() == 0 ) then

! BARRIER ID 0
!$OMP BARRIER

!$OMP SIMD             
             do j_1=1,n
                v_x(j_1) = ZERO
             end do! j_1

! BARRIER ID 1
!$OMP BARRIER

         end if

       end if

       t1 = omp_get_wtime()
!$OMP MASTER
             ttt6(5) = ttt6(5) + (t1 - t0)
!$OMP END MASTER


       return
       end subroutine ! tred1_vorhto

