#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <cuda_runtime.h>
#include <cuda.h>
#include <cublas.h>

#include "misc_gpu.hpp"

extern "C"
cudaStream_t ASPEN_get_Stream(void);


__global__ void
t6xx_gpu( const int L, double * beta_, double * u_x, double * v_x )
{
   const int myid   = threadIdx.x + 1;

#define	u_x(x)		u_x[(x)-1]
#define	v_x(x)		v_x[(x)-1]

/*
   Scal(l, beta, &v(1), 1);
   const T gamma = Dot(l, &u(1), 1, &v(1), 1) * beta * HALF;
   Axpy(l, gamma, &u(1), 1, &v(1), 1);
*/

   if ( blockIdx.x == 0 ) {

      const double beta = *beta_;

      double gamma = 0.;
      for(int k=myid; k<=L; k+=blockDim.x) {
         const double t = beta*v_x(k);
         gamma += t * u_x(k);
      } sum_over_warp_masters(gamma);

      gamma *= beta; gamma /= 2;
      for(int k=myid; k<=L; k+=blockDim.x) {
         v_x(k) = fma(gamma, u_x(k), beta*v_x(k));
      }

   }
}

static int    THREADS_dynamic = 0;
static int    BLOCKS_dynamic  = 0;
static size_t SHMEM_required  = 0;

extern "C" void
t6xx( const int L, double * beta, double * dev_ux, double * dev_vx, cudaStream_t stream )
{
   const int THREADS_max = (32*32);

   if ( THREADS_dynamic <= 0 ) {
      THREADS_dynamic = THREADS_max;
      while ( true ) {
         SHMEM_required = get_shmem_size(THREADS_dynamic);
         int numBlocks;
         cudaOccupancyMaxActiveBlocksPerMultiprocessor( &numBlocks, t6xx_gpu,
                         THREADS_dynamic,
                         SHMEM_required );
         if ( numBlocks > 0 || THREADS_dynamic <= 0 ) break;
         THREADS_dynamic -= 32;
      }

      struct cudaDeviceProp deviceProp;
      cudaGetDeviceProperties ( &deviceProp, 0 );
      BLOCKS_dynamic = deviceProp.multiProcessorCount;
   }

   t6xx_gpu <<< BLOCKS_dynamic, THREADS_dynamic, SHMEM_required, stream >>>
      ( L, beta, dev_ux, dev_vx );
}

