       subroutine trbakwyx(n, nvec, a, lda, z, ldz, d, e, m0)
       use OMP_LIB
       implicit NONE

       integer, intent(IN)    :: n, nvec, lda, ldz, m0
       real(8), intent(INOUT) :: a(*)
       real(8), intent(INOUT) :: z(*)
       real(8), intent(INOUT) :: d(*)
       real(8), intent(IN)    :: e(*)

       real(8) , pointer      :: w(:), v(:)
       real(8) , pointer      :: ss(:)
       integer                :: i_v, i_s

       integer, parameter     :: nsm = 256

       integer                :: N_GPU, N_GPU_a, N_GPU_b
       integer                :: m, na, nm, kx, ierr
       integer                :: i, j
       real(8)                :: d1, d2
*
       character(len=255)     :: gpu_cpu_rate
*
       integer, external      :: cuda_alloc
       external               :: cuda_free

       integer(8)             :: i8_
       integer(8)             :: devPtr_Root
       integer(8)             :: WorkLEN_GPU
       integer(8)             :: devPtrZ, devPtrV, devPtrSS
       integer(8)             :: LEN_Z, LEN_V, LEN_SS
*
       include 'CSTAB.h'
*
!       call cublas_init( )
*
       m = MAX(MIN(nsm,m0),1)
*
       na   = n +MOD(n+1,2)
       nm  = lda
!       call CSTAB_get_optdim(lda,9,16*4,16*6,nm)
!       na   = nm
*
       allocate(
     &          v(1:MAX(nm*m,n)),
     &          ss(1:na*m+nsm*nsm),
     &          stat=ierr)
       if(ierr/=0)then
         print*,"Memory allocation error [TBK]."
         stop
       endif
       v(:)=0D0
       ss(:)=0D0
*
       N_GPU_a=29; N_GPU_b=32
*
       call getenv( "GPU_CPU_RATE", gpu_cpu_rate )
       if ( gpu_cpu_rate(1:1) /= ' ' ) then
          read(gpu_cpu_rate,*) N_GPU_a, N_GPU_b
          N_GPU_b = N_GPU_b + N_GPU_a
       end if
*
       N_GPU=(NVEC*N_GPU_a)/N_GPU_b
!       N_GPU = NVEC
!       print*," :: GPU / CPU = ", N_GPU, NVEC-N_GPU
*
       if ( N_GPU > 0 ) then

          LEN_Z  = int(ldz,8)*int(N_GPU,8)
          LEN_V  = int(nm,8)*int(m,8)
          LEN_SS = int(nm,8)*int(m,8) + int(nsm,8)*int(nsm,8)

          WorkLEN_GPU = LEN_Z + LEN_V + LEN_SS
          call cuda_sync( 1 )
          ierr = cuda_alloc( devPtr_Root, 8*WorkLEN_GPU )
          call cuda_sync( 1 )
          if ( ierr /= 0 ) then
             print*,"Device Memory allocation fault [TBK]"
             stop
          end if

          devPtrZ  = devPtr_Root
          devPtrV  = devPtrZ     + LEN_Z * 8
          devPtrSS = devPtrV     + LEN_V * 8

       else

          devPtrZ = 0
          devPtrV = 0
          devPtrSS = 0

       end if
*
*
       d1 = OMP_Get_wtime()
!$OMP PARALLEL
       call trbakwy_body( n, nvec,
     $      a(1), lda,
     $      z(1), devPtrZ, ldz,
     $      e(1), d(1),
     $      v(1:), devPtrV, nm,
     $      m,
     $      ss(1:), devPtrSS, na,
     $      N_GPU )
!$OMP END PARALLEL
       d2 = OMP_Get_wtime()
*
*
       if ( N_GPU > 0 ) then
       call cuda_sync( 1 )
       call cuda_free(devPtr_Root)
       call cuda_sync( 1 )
       end if
*
*
       deallocate( v )
       deallocate( ss )
*
!       call cublas_shutdown( )

!       print*,"TRBAKWY",d2-d1
!       call flush(6)
*
       return
       end subroutine
*
*
*
       subroutine trbakwy_body(
     &             n, nvec,
     &             a, lda,
     &             z, devPtrZ, ldz,
     &             e,d,
     &             v,devPtrV,nm,
     &             m,
     &             ss, devPtrSS, nss,
     &             N_GPU)
       use OMP_LIB
       implicit NONE

       integer, intent(IN)    :: n, nvec, lda, ldz, nm, m
       integer, intent(IN)    :: nss, N_GPU
       real(8), intent(INOUT) :: a(1:lda,*)
       real(8), intent(OUT)   :: z(1:ldz,*)
       real(8), intent(IN)    :: e(1:n)
       real(8), intent(OUT)   :: d(1:n)
       real(8), intent(OUT)   :: v(1:nm,*)
       real(8), intent(OUT)   :: ss(*)

       integer(8), intent(IN) :: devPtrZ, devPtrV, devPtrSS

       integer, parameter     :: nsm = 256
*
       integer                :: j_1, j_2, j_3, j_4
       integer                :: i,j,k, nx
       real(8)                :: s0
       real(8)                :: d1,d2,dx,dy
       integer(8)             :: mlen
*
       integer, external      :: cublas_set_matrix, cublas_get_matrix
       integer, external      :: cublas_set_vector
       integer, external      :: cublas_set_matrix_async
*
#if TIMER==1
!$OMP MASTER
       d1=OMP_Get_wtime()
       dx = 0.0; dy = 0.0
!$OMP END MASTER
#endif
*
       nx = MIN(MOD(n-2+1,m)+2-1,n)
*-
!$OMP MASTER
       if ( N_GPU > 0 ) then
          mlen = int(ldz,8)*int(N_GPU,8)
          call cuda_sync( 1 )
          call cuda_set_vector ( mlen, int(8,8), z, devPtrZ )
          call cuda_sync( 1 )
       end if
!$OMP END MASTER
*--
       if ( omp_get_thread_num() ==
     &        MIN( 1, omp_get_num_threads()-1 ) ) then
*-
         if ( m >= 2 ) then
            i=1
               do j=2,nx-1
                  a(i+j-1:i+nx-2,i+j-1) = 0.0D+00
               end do
            do i=nx+1, n, m
               do j=1,m-1
                  a(i+j-1:i+m-2,i+j-1) = 0.0D+00
               end do
            end do
         end if
*-
       end if
*--
          call trbakwy_body2(
     $         nvec,
     $         z(1,1), devPtrZ, ldz,
     $         a(1,2), devPtrV, nm,
     $         nx-1, 2,
     $         ss, devPtrSS, nss,
     $         dx, dy, N_GPU)
*-
       do i=nx+1, n, m
*-
          call trbakwy_body2(
     $         nvec,
     $         z(1,1), devPtrZ, ldz,
     $         a(1,i), devPtrV, nm,
     $         m, i,
     $         ss, devPtrSS, nss,
     $         dx, dy, N_GPU)
*-
       end do
*--
!$OMP BARRIER
*-
!$OMP MASTER
       if ( N_GPU > 0 ) then
          mlen = int(ldz,8)*int(N_GPU,8)
          call cuda_sync( 1 )
          call cuda_get_vector ( mlen, int(8,8), devPtrZ, z )
          call cuda_sync( 1 )
       end if
!$OMP END MASTER
*-
#if TIMER==1
!$OMP BARRIER
*
!$OMP MASTER
       d2=OMP_Get_wtime()

       print*,"TRBAK=",(d2-d1)
       print*,"   ",(2d0*n*n*n)/(d2-d1)*1d-9,"GFLOPS"
       print*," 1 ",(1d0*n*n*n)/(dx)*1d-9,"GFLOPS"
       print*," 2 ",(1d0*n*n*n)/(dy)*1d-9,"GFLOPS"

       call flush(6)
!$OMP END MASTER
*
!$OMP BARRIER
#endif
*
*
       return
       end subroutine

