       subroutine trbakwy_body2(
     $            nvec,
     $            z, devPtrZ, nmz,
     $            v, devPtrV, nm,
     $            m, i,
     $            ss, devPtrSS, nss,
     $            dx, dy, N_GPU)
       use OMP_LIB
       implicit NONE
*
       integer, intent(IN)    :: nvec, nmz, nm, m, i, nss
       integer, intent(IN)    :: N_GPU
       real(8), intent(INOUT) :: z(nmz,*)
       real(8), intent(INOUT) :: v(nm,*)
       real(8), intent(INOUT) :: ss(*)
       integer(8), intent(IN) :: devPtrZ, devPtrV, devPtrSS
       real(8), intent(INOUT) :: dx, dy
       integer, parameter     :: nsm = 256
*
       integer                :: i_1, i_2, i_3, i_4
       integer                :: j_1, j_2, j_3, j_4
       integer                :: i_0, m_0, k, mm
       integer(8)             :: idxSS
       real(8)                :: ds, de
       external               :: cublas_dscal
*
!$OMP BARRIER
*-
! SM:= 0
! SS:= 0
       if ( omp_get_num_threads()==1 ) then
          do i_1=1,nss*m+nsm*nsm
             ss(i_1) = 0D0
          end do
       else if ( omp_get_thread_num()>=1 ) then
          j_1 = nss*m+nsm*nsm
          j_2 = (j_1-1)/(omp_get_num_threads()-1)+1
          i_2 = j_2*(omp_get_thread_num()-1)+1
          i_3 = MIN(i_2+j_2-1, j_1)
          do i_1=i_2,i_3
             ss(i_1) = 0D0
          end do
       end if
*
          i_2 = 1
          i_3 = nvec
*
          j_2 = 1   
          j_3 = i+m-2
*
*-------------
!$OMP BARRIER
*-------------
*
! SM:= SM+lower(V^TV)
! SS(1:m,1:n):= SS+V(j_2:j_3,1:m)^T*Z(j_2:j_3,1:n)
#if TIMER==1
          ds=OMP_Get_wtime()
#endif
          call trbk1(
     $               z(1,1), devPtrZ, nmz,
     $               v(1,1), devPtrV, nm,
     $               m,
     $               ss(1+nsm*nsm), devPtrSS,
     $               ss(1),
     $               i_2,i_3, j_2,j_3,
     $               N_GPU )
#if TIMER==1
!$OMP BARRIER
          de=OMP_Get_wtime()
!$OMP MASTER
          dx = dx + (de-ds)
!$OMP END MASTER
#endif
*-
*------------
*-
!$OMP BARRIER
! V:= V*(I-SM)^{-1}
! Z(j_2:j_3,1:n):= Z + V(j_2:j_3,1:m)*SS(1:m,1:n)
#if TIMER==1
          ds=OMP_Get_wtime()
#endif
          call trbk2(
     $                z(1,1), devPtrZ, nmz,
     $                v(1,1), devPtrV, nm,
     $                m,
     $                ss(1+nsm*nsm), devPtrSS,
     $                ss(1),
     $                i_2,i_3, j_2,j_3,
     $                N_GPU )

#if TIMER==1
!$OMP BARRIER
          de=OMP_Get_wtime()
!$OMP MASTER
          dy = dy + (de-ds)
!$OMP END MASTER
#endif
*-

       return
       end subroutine ! trbakwy_body2
*-
       subroutine trbk1(z, devPtrZ, nmz,
     $            v, devPtrV, nm, m,
     $            ss, devPtrSS, sm, i_2,i_3,j_2,j_3, N_GPU)
       use OMP_LIB
       implicit double precision(a-h,o-z),integer(i-n)
*
       integer, intent(IN)    :: nmz, nm, m, N_GPU
       real(8), intent(INOUT) :: z(nmz,*),v(nm,*)
       integer(8), intent(IN) :: devPtrZ, devPtrV, devPtrSS
       real(8), intent(INOUT) :: ss(m,*)
       integer, parameter     :: nsm = 256 
       real(8), intent(INOUT) :: sm(nsm,nsm)
*
       integer, external      :: cublas_set_matrix
       external               :: dzero, DGEMM, CUBLAS_DGEMM
*
       integer                :: local_rank, local_size
       integer                :: blk_size1, blk_size2
*
       integer                :: BLAS_CHUNK1 = 128
       integer                :: BLAS_CHUNK2 = 64

       integer(8), parameter  :: eSize = 8
       integer(8)             :: mlen
*
          j_5 = j_3 - j_2 + 1
*
!$OMP MASTER
          if ( j_5 > 0 .AND.  N_GPU > 0 ) then

             mlen = int(nm,8)*int(m-1,8)
     &             +int(j_5,8)
             call cuda_sync( -1 )
             call cuda_set_vector( mlen, eSize, v, devPtrV )
             call cuda_sync( 1 )

!             call CUBLAS_DGEMM('T','N',
!     &               m, m, j_5,
!     &              -1.0D+00, devPtrV, nm,
!     &                        devPtrV, nm,
!     &               0.0D+00, devPtrSS, nsm)
             call CUBLAS_DSYRK('L','T',
     &               m, j_5,
     &              -1.0D+00, devPtrV, nm,
     &               0.0D+00, devPtrSS, nsm)

             call cuda_sync( 1 )
             call mul_diag_gpu( m, devPtrSS, nsm+1 )
             call cuda_sync( -1 )
             if ( i_3 /= N_GPU ) then
                mlen = int(m,8)*int(nsm,8)
                call cuda_sync( -1 )
                call cuda_get_vector( mlen, eSize, devPtrSS, sm )
             end if

             idxZ =int(i_2-1,8)*int(nmz,8)*8
             idxSS=int(nsm,8)*int(nsm,8)*8
             call CUBLAS_DGEMM('T','N',
     &               m, N_GPU, j_5,
     &               1.0D+00, devPtrV,  nm,
     &                        devPtrZ +idxZ,  nmz,
     &               0.0D+00, devPtrSS+idxSS, m)

             if ( i_3 /= N_GPU ) then
                call cuda_sync( 1 )
             end if

          end if
!$OMP END MASTER
*
       local_size = omp_get_num_threads()
       local_rank = omp_get_thread_num()
       ll_size = local_size
       ll_rank = local_rank
       if ( local_size > 1 ) then
          ll_size = ll_size - 1
          ll_rank = ll_rank - 1
       endif
*-
       if ( N_GPU == 0 ) then
       if ( local_rank > 0 .OR. local_size == 1 ) then
*-
       if ( j_5 > 0 ) then

          ii_step = 0
          do i_1 = 1, m, BLAS_CHUNK1
          do j_1 = 1, m, BLAS_CHUNK2

             ii_3 = MIN(m, i_1+BLAS_CHUNK1-1)
             jj_3 = MIN(m, j_1+BLAS_CHUNK2-1)

             blk_size1 = ii_3 - i_1 + 1
             blk_size2 = jj_3 - j_1 + 1

          if ( blk_size1 > 0 .AND. blk_size2 > 0 ) then
             if ( i_1+blk_size1 >= j_1+blk_size2 ) then

                ii_2 = MIN(MAX(i_1,j_1),ii_3)
                blk_size1 = ii_3 - ii_2 + 1

                if ( MOD(ii_step, ll_size) == ll_rank ) then

                call DGEMM('T','N',
     &               blk_size1, blk_size2, j_5,
     &              -1.0D+00, v(j_2, ii_2), nm,
     &                        v(j_2, j_1),  nm,
     &               0.0D+00, sm(ii_2,j_1), nsm)

                end if

                ii_step = ii_step + 1

             end if
          end if

          end do
          end do

       end if
*-
       end if
       end if
*-
          j_5 = j_3 - j_2 + 1
          ii_2 = N_GPU+1

       if ( local_rank > 0 .OR. local_size == 1 ) then

          i_0 = (i_3-ii_2) / ll_size + 1
          i_4 = i_0 * ll_rank + ii_2
          i_5 = MIN(i_0, i_3-i_4+1)

          if ( j_5 > 0 .AND. i_5 > 0 ) then
             call DGEMM('T','N',
     &               m, i_5, j_5,
     &               1.0D+00, v (j_2    ,1  ), nm,
     &                        z (j_2    ,i_4), nmz,
     &               0.0D+00, ss(1      ,i_4), m)
          end if

       end if

!$OMP BARRIER
*-
       if ( N_GPU == 0 ) then
!$OMP DO
       do j_1=1,m
         if ( sm(j_1,j_1) == 0D0 ) then
           sm(j_1,j_1) = 1D0
         else
           sm(j_1,j_1) = sm(j_1,j_1) * 5D-1
         end if
       end do
!$OMP END DO
       end if

       return
       end subroutine ! trbk1
*-
       subroutine trbk2(
     $            z, devPtrZ, nmz,
     $            v, devPtrV, nm, m,
     $            ss, devPtrSS, sm, i_2, i_3, j_2, j_3, N_GPU )
       use OMP_LIB
       implicit double precision(a-h,o-z),integer(i-n)
*
       integer, intent(IN)    :: nmz, nm, m, N_GPU
       real(8), intent(INOUT) :: z(nmz,*)
       real(8), intent(INOUT) :: v(nm,*)
       real(8), intent(INOUT) :: ss(m,*)
       integer(8), intent(IN) :: devPtrZ, devPtrV, devPtrSS
       integer, parameter     :: nsm = 256 
       real(8), intent(INOUT) :: sm(nsm,*)
       integer, intent(IN)    :: i_2, i_3, j_2, j_3
*
       integer, external      :: cublas_set_matrix
       external               :: DGEMM, CUBLAS_DGEMM
*
       real(8)                :: ds, de
       integer                :: local_rank, local_size
       integer                :: ll_rank, ll_size



       local_size = omp_get_num_threads()
       local_rank = omp_get_thread_num()
       ll_size = 1
       ll_rank = 0
       if ( local_size > 1 ) then
          ll_size = local_size - 1
          ll_rank = local_rank - 1
       endif
*
       if ( local_rank > 0 .OR. local_size == 1 ) then
*-
          j_5 = (j_3-j_2) / ll_size + 1
          j_5 = ((j_5-1)/2+1)*2
          j_4 = j_2 + j_5 * ll_rank

          j_5 = MIN(j_5, j_3-j_4+1)
          j_6 = j_4
          j_7 = j_6+j_5-1

       if ( j_5 > 0 ) then

          call DTRSM('R','L','N','N',
     &               j_5, m, 1.0D+00, sm(1,1), nsm, v(j_6,1), nm)

       end if
         

       end if
*-
!$OMP BARRIER

          j_5 = j_3 - j_2 + 1

! Z[1:i_7]=Z[1:i_7]+V*SS[1:i_7]

!$OMP MASTER
          if ( j_5 > 0 .AND. N_GPU > 0 ) then

             idxV =(int(j_2-1,8)
     &             +int(1-1,8)*int(nm,8))*8
             idxSS=(int(1-1,8)
     &             +int(1-1,8)*int(m,8)
     &             +int(nsm,8)*int(nsm,8))*8
             idxZ =(int(j_2-1,8)
     &             +int(i_2-1,8)*int(nmz,8))*8

             call CUBLAS_DTRSM('R','L','N','N',
     &               j_5, m, 1.0D+00, devPtrSS, nsm, devPtrV+idxV, nm)

             call CUBLAS_DGEMM('N','N',
     &               j_5, N_GPU, m,
     &               1.0D+00, devPtrV +idxV,  nm,
     &                        devPtrSS+idxSS, m,
     &               1.0D+00, devPtrZ +idxZ,  nmz)

          end if
!$OMP END MASTER

       if ( local_rank > 0 .OR. local_size == 1 ) then

          j_5 = j_3 - j_2 + 1

          if ( j_5 > 0 ) then

          ii_2 = N_GPU+1

          i_0 = (i_3-ii_2) / ll_size + 1
          i_4 = i_0 * ll_rank
          i_5 = MIN(i_0, i_3-(ii_2+i_4)+1)

          if ( i_5 > 0 ) then
             call DGEMM('N','N',
     &               j_5, i_5, m,
     &               1.0D+00, v (j_2    ,1       ), nm,
     &                        ss(1      ,ii_2+i_4), m,
     &               1.0D+00, z (j_2    ,ii_2+i_4), nmz)
          end if

          end if

       end if

*-
*
       return
       end subroutine ! trbakwy_body2

