       subroutine tred1(n,a,nma0,d_out,e_out,m0)
       implicit NONE

       integer                :: n, nma0, m0
       integer                :: nm, m
       real(8)                :: a(1:nma0,*)
       real(8)                :: d_out(1:n), e_out(1:n)

       include 'trd.h'
       include 'CSTAB.h'


          nm  = nma0
          m   = m0

          call tred1_(a, d_out, e_out, n, nm, m)


       return
       end subroutine ! tred1

       subroutine tred1_(a,d_out,e_out,n,nm,m)
       implicit NONE

       integer                :: n, nm, nmx, nv, nvx, m
       real(8)                :: a(nm,*)
       real(8)                :: d_out(1:n)
       real(8)                :: e_out(1:n)

       real(8), pointer       :: work(:)

       include 'trd.h'
       include 'CSTAB.h'
       include 'param.h'

       real(8)                :: d1,d2
       real(8)                :: d3,d4
       integer                :: nm_orig, n1, n2, n3
       integer                :: i,j,k
       integer                :: ierr, kx

       integer(8)             :: devPtr_Root
       integer(8)             :: WorkLEN_GPU
       integer(8)             :: LEN_A, LEN_UX, LEN_VX, LEN_WK


          LEN_A  = int(nm,8)*int(n,8)
          LEN_UX = int(nm,8)*int(m,8)
          LEN_VX = int(nm,8)*int(m,8)
          LEN_WK = int(2,8)*int(m,8)

          WorkLEN_GPU = LEN_A + LEN_UX + LEN_VX + LEN_WK
          call cuda_sync( 1 )
          ierr = cuda_alloc( devPtr_Root, 8*WorkLEN_GPU )
          call cuda_sync( 1 )
          if ( ierr /= 0 ) then
             print*,"Device Memory allocation fault [TRD]"
             stop
          endif

          devPtr_A  = devPtr_Root
          devPtr_UX = devPtr_A    + LEN_A  * 8
          devPtr_VX = devPtr_UX   + LEN_UX * 8
          devPtr_UY = devPtr_VX   + LEN_VX * 8

          allocate( work(1:2*m+nm*m), stat=ierr )
          if ( ierr /= 0 ) then
             print*,"Memory allocation error [TRD]."
             stop
          end if

!$OMP PARALLEL
          call trd_body(a, nm, d_out, e_out, n, m,
     &                  work(1:),
     &                  work(1+nm*m:),
     &                  work(1+nm*m+m:)
     &                  )
!$OMP END PARALLEL

          call cuda_sync( 1 )
          call cuda_free(devPtr_Root)
          call cuda_sync( 1 )

          deallocate(work)

       return
       end subroutine ! tred1_

       subroutine trd_body(a, nm, d_out, e_out, n, m_orig,
     &               v_x, u_t, v_t)
       use OMP_LIB
       implicit NONE

       integer                :: n, nm, m_orig
       real(8)                :: a(1:nm,*)
       real(8)                :: d_out(1:n), e_out(1:n)
       real(8)                :: u_t(1:m_orig), v_t(1:m_orig)
       real(8)                :: v_x(1:nm,*)

       integer, parameter     :: MBAND = 1

       real(8), save          :: beta
       real(8)                :: t1, t2, t3
       integer                :: i, j, k, l

       integer                :: y_root, y_pos
       integer                :: x_root, x_pos
       integer                :: i_1, i_2, i_3, i_4
       integer                :: j_1, j_2, j_3, j_4
       integer                :: k_1, k_2, k_3, k_4
       integer                :: l_1, l_2, l_3, l_4

       integer                :: n_left, m_block, m, mb
       integer                :: i_block, i_base
       logical                :: is_the_master, is_1st_team
       logical                :: is_2nd_master, is_2nd_team

       integer                :: ii

       include 'trd.h'
       include 'param.h'

#define TIMER   0

#if TIMER==1
       real(8)      t2_reduce1, t2_reduce2
       common /t2/  t2_reduce1, t2_reduce2


       real(8)                :: d1,d2,dd(100)


       t2_reduce1 = ZERO
       t2_reduce2 = ZERO

       dd=0
#endif
!
! initialization
!
#if TIMER==1
!$OMP MASTER
       print*,"NUM.OF.THREADS=",omp_get_num_threads()
       call flush(6)
!$OMP END MASTER
#endif


       is_the_master = (omp_get_thread_num() == 0)
       is_1st_team   = .not. is_the_master
       is_2nd_master = (omp_get_thread_num() ==
     &                  MIN( 1, omp_get_num_threads()-1 ))
       is_2nd_team   = (omp_get_num_threads() == 1 .OR. is_1st_team)


       call tred1_init( a(1,1), nm, n, d_out(1) )

!$OMP BARRIER

       mb      = MIN(m_orig, n)
       m_block = ((n-MOD(n,MBAND))-1)/mb+1

       do i_block=m_block,MAX(1,3*(2-mb)),-1

!$OMP BARRIER

          i_base = (i_block-1)*mb
          m      = MIN(mb,n-i_base)

!$OMP MASTER
          if ( i_block /= m_block ) then

             ! Panel u_x is mapped the same as a(1,i_base+1:i_base+m)
             call tred1_load( a(1,i_base+1), nm, m, i_base, i_block )

          end if
!$OMP END MASTER
          !

          k_2 = m
          k_3 = MAX(1,3*(2-i_block))
          beta = ZERO

          do k_1=k_2,k_3,-1; k_4=k_1-MBAND+1

             i = i_base+k_1

             if ( k_1==k_2 .OR. is_1st_team ) then
!
! when initial (k_1==k_2) => all threads pass
! or other case           => non-0 pass
!
! this corresponds to the barrier before cublas_set_vector in t6
!
! BARRIER ID : 0
!$OMP BARRIER
!
! all threads pass
!
             end if

             if ( is_2nd_master ) then
!
! 1 or 0 pass
!
                if ( k_1 /= k_2 ) then

#if TIMER==1
                   d1=OMP_Get_wtime()
#endif
!
! u_x':= u_x-uv-vu
!
                   call tred1_local_2update0(
     &                  a(1,i_base+k_4), nm,
     &                  a(1,i_base+k_4+1), v_x(1,k_4+1),
     &                  beta,
     &                  i_base, i+1 )

#if TIMER==1
                   d2=OMP_Get_wtime()
                   dd(6)=dd(6)+(d2-d1)
#endif

                end if
!
! 0 or 1 pass
!
             end if
!
! all threads pass
!
             if ( k_1==k_2 .OR. is_1st_team ) then
!
! when initial (k_1==k_2) => all threads pass
! or other case           => non-0 pass
!
! this corresponds to the barrier after cublas_set_vector in t6
!
! BARRIER ID : 1
!$OMP BARRIER
!
             end if

#if TIMER==1
!$OMP MASTER
             d1=OMP_Get_wtime()
!$OMP END MASTER
#endif
!
! v:=Au
!
! all threads pass
!
             call tred1_aux(
     &               a(1,1), nm,
     &               a(1,i_base+k_4), v_x(1,k_4),
     &               i, i_base, m,
     &               d_out(1), e_out(1), beta )
!
! all threads pass
!
             if ( is_2nd_team ) then
                if ( k_1 < k_2 ) then

                   call tred1_local_2update(
     &                  a(1,i_base+1), nm,
     &                  a(1,i_base+k_4+1), v_x(1,k_4+1),
     &                  i_base, i-1+1, i+1)

                end if
             end if
!
! all threads pass
!
#if TIMER==1
!$OMP MASTER
             d2=OMP_Get_wtime()
             dd(2)=dd(2)+(d2-d1)
!
! v=v-(UV+VU)u
!
!
! v':= v-((u,v)/2|u|^2)u
!
             d1=OMP_Get_wtime()
!$OMP END MASTER
#endif

             call tred1_vo(
     &               a(1,i_base+k_4), v_x(1,k_4),
     &               a(1,i_base+1), v_x(1,1), nm,
     &               u_t(1), v_t(1), beta,
     &               i, i_base, m)

#if TIMER==1
!$OMP MASTER
             d2=OMP_Get_wtime()
             dd(4)=dd(4)+(d2-d1)
!$OMP END MASTER
#endif

          end do! k_1

          if ( is_1st_team ) then
             if ( k_2 >= k_3 ) then
!
! this corresponds to the barrier before cublas_set_vector in t6
!
! BARRIER ID 0
!$OMP BARRIER
!
! this corresponds to the barrier after cublas_set_vector in t6
!
! BARRIER ID 0
!$OMP BARRIER
             end if
          end if

          if ( is_2nd_team ) then
             if ( i_base == 0 ) then

                k_1=k_3; k_4=k_1-MBAND+1
                i = k_1

#if TIMER==1
                d1=OMP_Get_wtime()
#endif
!
! w':= w-uv-vu
!
                call tred1_local_2update(
     &                  a(1,i_base+1), nm,
     &                  a(1,i_base+k_4), v_x(1,k_4),
     &                  i_base, i, i)

#if TIMER==1
                d2=OMP_Get_wtime()
                dd(5)=dd(5)+(d2-d1)
#endif
             end if

          end if

!$OMP BARRIER

#if TIMER==1
!$OMP MASTER
          d1=OMP_Get_wtime()
!$OMP END MASTER
#endif
!
! A:=A-v^Tu-uT^v
!
!$OMP MASTER
          if ( i_block > 1 ) then
             call tred1_2update(
     &               a(1,1), nm,
     &               a(1,i_base+1),v_x(1,1),
     &               m, i_base )
          end if
!$OMP END MASTER

!$OMP BARRIER

#if TIMER==1
!$OMP MASTER
          d2=OMP_Get_wtime()
          dd(3)=dd(3)+(d2-d1)
!$OMP END MASTER
#endif

       end do! i_block

!$OMP BARRIER

!$OMP MASTER

       call tred1_final(a(1,1), nm, n, d_out(1), e_out(1), mb)

#if TIMER==1
       print*,"calc (u,beta)    ",dd(1)
       print*,"mat-vec (Au)     ",dd(2),dble(n)**3*2d-9/3/dd(2)
       print*,"  COMM1/2        ",t2_reduce1,t2_reduce2
       print*,"2update (A-uv-vu)",dd(3),dble(n)**3*2d-9/3/dd(3)
       print*,"v=v-(UV+VU)u     ",dd(4)
       print*,"UV post reduction",dd(5)

       call flush(6)
#endif

!$OMP END MASTER


       return
       end subroutine ! trd_body

